import logging

from tracker_exporter.models.base import Base
from tracker_exporter.models.enums import (
    TrackerChangelogEvents,
    TrackerWorkflowTypes,
)
from tracker_exporter.services.monitoring import DogStatsdClient
from tracker_exporter.utils.helpers import (
    calculate_time_spent,
    string_normalize,
    validate_resource,
    to_simple_datetime,
    to_snake_case,
    to_human_time,
)
from tracker_exporter.defaults import CLOSED_ISSUE_STATUSES

logger = logging.getLogger(__name__)
monitoring = DogStatsdClient()


class TrackerIssueMetric(Base):
    """This object represents a issue metrics for TrackerIssue object."""
    def __init__(self,
                 issue_key: str,
                 status_name: str,
                 status_transitions_count: int,
                 duration: int,
                 busdays_duration: int,
                 last_seen: str):

        self.issue_key = issue_key
        self.status_name = status_name
        self.status_transitions_count = status_transitions_count
        self.duration = duration
        self.human_readable_duration = to_human_time(self.duration)
        self.busdays_duration = busdays_duration
        self.human_readable_busdays_duration = to_human_time(self.busdays_duration)
        self.last_seen = last_seen


class TrackerIssue(Base):
    """This object represents a issue from Yandex.Tracker."""

    def __init__(self, issue: object) -> None:
        self._issue = issue
        self._transform(self._issue)

    def _transform(self, issue: object) -> None:
        """Formation of a task object based on its metadata."""
        self.queue = issue.queue.key
        self.issue_key = issue.key
        self.title = string_normalize(issue.summary)
        self.issue_type = to_snake_case(validate_resource(issue.type, "name"))
        self.priority = validate_resource(issue.priority, "name")
        self.assignee = validate_resource(issue.assignee, "email")
        self.author = validate_resource(issue.createdBy, "email")
        self.status = to_snake_case(validate_resource(issue.status, "name"))
        self.resolution = to_snake_case(validate_resource(issue.resolution, "name"))
        self.tags = issue.tags or []
        self.components = [c.name for c in issue.components if issue.components]
        self.created_at = to_simple_datetime(issue.createdAt, date_only=True)
        self.updated_at = to_simple_datetime(issue.updatedAt, date_only=True)
        self.deadline = validate_resource(issue, "deadline")
        self.resolved_at = to_simple_datetime(issue.resolvedAt, date_only=True)
        self.start_date = validate_resource(issue, "start")
        self.end_date = validate_resource(issue, "end")
        self.story_points = validate_resource(issue, "storyPoints") or 0
        self.sprints = [s.name for s in issue.sprint if issue.sprint]
        self.parent_issue_key = validate_resource(issue.parent, "key", low=False)
        self.epic_issue_key = validate_resource(issue.epic, "key", low=False)
        self.is_subtask = True if any((self.parent_issue_key,)) else False
        self.is_closed = True if self.status in CLOSED_ISSUE_STATUSES else False
        self.is_resolved = True if self.resolution is not None else False
        self.qa_engineer = validate_resource(issue.qaEngineer, "email")

    @monitoring.send_count_metric("issues_total_processed_count", 1, tags=["source:issues"])
    def metrics(self) -> list:
        """
        All metrics are based on status change events in the task history.
        The method has the ability to filter only the necessary statuses
        passed in the argument.

        The metric of being in the status is considered
        only after the end of being in the calculated status.

        For example, the task has moved from the status "Open"
        to the status "In progress", in this case only the metric
        for "Open" will be considered.
        As soon as the status "In progress" is changed to any other,
        it will be calculated as a metric for "In progress".

        In other words, the current status of the task will not be
        calculated.
        """
        metrics_storage = {}

        for event in self._issue.changelog:
            if event.type == TrackerChangelogEvents.ISSUE_WORKFLOW:
                logger.debug(f"Issue workflow fields found: {event.fields}")
                worklow_type = event.fields[0].get("field").id
                # Keep only status transition events, drop otherwise
                if worklow_type != TrackerWorkflowTypes.TRANSITION:
                    logger.debug(f"Skipping {event.fields[0].get('field').id} for {self.issue_key}")
                    continue

                status = to_snake_case(event.fields[0].get("from").name.lower())
                event_start_time  = event.fields[1].get("from")
                event_end_time = event.fields[1].get("to")

                # Custom logic for calculating the completion date of the task,
                # because not everyone uses resolutions, sadly
                transition_status = to_snake_case(event.fields[0].get("to").name.lower())
                if transition_status in CLOSED_ISSUE_STATUSES and self.status in CLOSED_ISSUE_STATUSES:
                    self.closed_at = to_simple_datetime(event_end_time, date_only=True)  # pylint: disable=W0201

                if event_start_time is None or event_end_time is None:
                    continue

                # Calculation of the time spent in the status
                start_time = to_simple_datetime(event_start_time)
                end_time = to_simple_datetime(event_end_time)
                total_status_time = calculate_time_spent(start_time, end_time)
                busdays_status_time = calculate_time_spent(start_time, end_time, busdays_only=True)
                try:
                    metrics_storage[status]["duration"] += total_status_time
                    metrics_storage[status]["busdays_duration"] += busdays_status_time
                    metrics_storage[status]["status_transitions_count"] += 1
                except (KeyError, AttributeError):
                    metrics_storage[status] = dict(
                        issue_key=self.issue_key,
                        status_name=status,
                        status_transitions_count=1,
                        duration=total_status_time,
                        busdays_duration=busdays_status_time,
                        last_seen=to_simple_datetime(event_end_time)
                    )

        logger.debug(f"Metrics for {self.issue_key}: {metrics_storage}")
        metrics = [TrackerIssueMetric(**metric) for _, metric in metrics_storage.items()]
        return metrics



class TrackerExporterError(Exception):
    pass


class ClickhouseError(TrackerExporterError):
    pass


class TrackerError(TrackerExporterError):
    pass


class NetworkError(TrackerExporterError):
    pass


class ExportError(TrackerExporterError):
    pass


class TimedOut(TrackerExporterError):
    def __init__(self):
        super().__init__("Timed out")



import time
import logging

from typing import Union, Tuple, List
from tracker_exporter.services.tracker import YandexTrackerClient
from tracker_exporter.services.clickhouse import ClickhouseClient
from tracker_exporter.services.monitoring import DogStatsdClient
from tracker_exporter.utils.helpers import to_human_time
from tracker_exporter.defaults import (
    CLICKHOUSE_HOST,
    CLICKHOUSE_HTTP_PORT,
    CLICKHOUSE_USER,
    CLICKHOUSE_PASSWORD,
    CLICKHOUSE_DATABASE,
    CLICKHOUSE_ISSUES_TABLE,
    CLICKHOUSE_ISSUE_METRICS_TABLE,
    TRACKER_BULK_CYCLE_TIME_ISSUES_LIMIT,
    EXCLUDE_QUEUES,
    MONITORING_ENABLED,
    TRACKER_ORG_ID,
    TRACKER_TOKEN,
    TRACKER_ISSUES_SEARCH_RANGE,
    TRACKER_ISSUES_SEARCH_QUERY
)

logger = logging.getLogger(__name__)
tracker = YandexTrackerClient(token=TRACKER_TOKEN, org_id=TRACKER_ORG_ID)
monitoring = DogStatsdClient(enabled=MONITORING_ENABLED)


class Exporter:
    # TODO: configure class instance
    # TODO: parse migration from sprint to sprint by changelog (field changed),
    # by default exported only last sprint (tracker logic)
    def __init__(self, clickhouse_client: ClickhouseClient = None):
        self.clickhouse = clickhouse_client or ClickhouseClient(
            host=CLICKHOUSE_HOST,
            port=CLICKHOUSE_HTTP_PORT,
            user=CLICKHOUSE_USER,
            password=CLICKHOUSE_PASSWORD
        )

    @monitoring.send_time_metric("issues_processing_time_seconds")
    def _bulk_issue_cycle_time(self,
                               query: str,
                               limit: int = 50) -> Tuple[List]:
        """Collects and transforms metrics for found tasks."""
        issues = []
        metrics = []
        issues_without_metrics = 0
        found_issues = tracker.search_issues(query=query, limit=limit)
        logger.info("Prepare cycle time metrics...")

        for tracker_issue in found_issues:
            try:
                issue, metric = tracker.issue_cycle_time(tracker_issue.key)
                if metric is None:
                    logger.debug(f"Ignore {tracker_issue.key} because metrics is empty")
                    issues_without_metrics += 1
                    issues.append(issue)
                else:
                    issues.append(issue)
                    for m in metric:  # pylint: disable=C0103
                        metrics.append(m)
            except Exception as exc:
                logger.exception(f"Issue {tracker_issue.key} can't be transformed, details: {exc}")

        monitoring.send_gauge_metric("issues_without_metrics", value=issues_without_metrics)
        logger.info(
            f"Total issues: {len(issues)}, total cycle time metrics: {len(metrics)}, "
            f"ignored issues with empty metrics: {issues_without_metrics}"
        )
        return issues, metrics


    def _upload_data_to_storage(self, payload: list, table: str) -> None:
        """Inserts a batch of data into the Clickhouse with deduplication."""
        logger.info(f"Inserting batch ({len(payload)} rows) to Clickhouse ({table})...")
        self.clickhouse.insert_batch(CLICKHOUSE_DATABASE, table, payload)

        logger.info(f"Optimizing table '{table}' for deduplication...")
        self.clickhouse.deduplicate(CLICKHOUSE_DATABASE, table)


    @monitoring.send_time_metric("cycle_time_total_processing_time_seconds")
    def cycle_time(self,
                   query: str = TRACKER_ISSUES_SEARCH_QUERY,
                   exclude_queues: Union[list, tuple] = EXCLUDE_QUEUES,
                   search_range: str = TRACKER_ISSUES_SEARCH_RANGE,
                   upload: bool = True) -> int:
        """Export issues cycle time and upload its to storage."""
        logger.info("Started processing issues...")
        if query:
            logger.warning("Arguments `excluded_queues`, `search_range` has no effect if a `query` is passed")
        queues = ", ".join([f"!{q}" for q in exclude_queues])
        _default_query = f"Queue: {queues} AND Updated: >= now() - {search_range}"
        search_query = TRACKER_ISSUES_SEARCH_QUERY or _default_query
        start_time = time.time()

        issues, metrics = self._bulk_issue_cycle_time(
            search_query,
            limit=TRACKER_BULK_CYCLE_TIME_ISSUES_LIMIT
        )

        if upload:
            self._upload_data_to_storage(issues, table=CLICKHOUSE_ISSUES_TABLE)
            self._upload_data_to_storage(metrics, table=CLICKHOUSE_ISSUE_METRICS_TABLE)
        else:
            logger.debug("Upload to Clickhouse is disabled")

        elapsed_time = time.time() - start_time
        logger.info(
            f"Processing issues completed. Elapsed time: {to_human_time(elapsed_time)}, "
            f"total tasks processed: {len(issues)}"
        )

        return len(issues) if upload else 0



import json
import logging

from typing import List, Dict, Union

from requests import Response
import requests

from tracker_exporter.errors import ClickhouseError, NetworkError, TimedOut
from tracker_exporter.utils.helpers import retry
from tracker_exporter.services.monitoring import DogStatsdClient
from tracker_exporter.models.enums import ClickhouseProto
from tracker_exporter.defaults import (
    CLICKHOUSE_PROTO,
    CLICKHOUSE_CACERT_PATH,
    CLICKHOUSE_SERVERLESS_PROXY_ID
)

logger = logging.getLogger(__name__)
monitoring = DogStatsdClient()


class ClickhouseClient:
    """This class provide simple facade interface for Clickhouse."""

    def __init__(self,  # pylint: disable=W0102
                 host: str,
                 port: int = 8123,
                 user: str = "default",
                 password: str = None,
                 proto: ClickhouseProto = CLICKHOUSE_PROTO,
                 cacert: str = CLICKHOUSE_CACERT_PATH,
                 serverless_proxy_id: str = CLICKHOUSE_SERVERLESS_PROXY_ID,
                 params: dict = {},
                 http_timeout: int = 10) -> None:

        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.proto = proto
        self.cacert = cacert
        self.serverless_proxy_id = serverless_proxy_id
        self.params = params
        self.timeout = int(http_timeout)
        self.headers = {"Content-Type": "application/json"}

        if self.proto == ClickhouseProto.HTTPS:
            assert self.cacert is not None

    @retry((NetworkError, TimedOut))
    def execute(self, query: str) -> Union[None, Response]:
        url = f"{self.proto}://{self.host}:{self.port}"

        if self.proto != ClickhouseProto.HTTPS:
            url += f"?user={self.user}"
            if self.password is not None:
                url += f"&password={self.password}"
        else:
            self.headers["X-Clickhouse-User"] = self.user
            self.headers["X-Clickhouse-Key"] = self.password

        if self.serverless_proxy_id:
            self.params["database"] = self.serverless_proxy_id

        if self.params:
            params = "&".join([f"{k}={v}" for k, v in self.params.items()])
            url += f"&{params}" if self.proto != ClickhouseProto.HTTPS else f"?{params}"

        try:
            if self.proto == ClickhouseProto.HTTPS:
                response = requests.post(
                    url=url, headers=self.headers, data=query,
                    timeout=self.timeout, verify=self.cacert
                )
            else:
                response = requests.post(
                    url=url, headers=self.headers, data=query, timeout=self.timeout
                )
        except requests.Timeout as exc:
            raise TimedOut() from exc
        except requests.ConnectionError as exc:
            raise NetworkError(exc) from exc
        except Exception as exc:
            logger.exception(
                f"Could not execute query in Clickhouse: {exc}"
            )
            raise ClickhouseError(exc) from exc
        else:
            if not response.ok:
                msg = (
                    f"Could not execute query in Clickhouse. "
                    f"Status: {response.status_code}. {response.text}"
                )
                logger.error(msg)
                raise ClickhouseError(msg)
        return response

    # TODO: add sort by partition key (i.e. `updated_at`) for best insert perfomance
    @monitoring.send_time_metric("clickhouse_insert_time_seconds")
    def insert_batch(self, database: str, table: str, payload: List[Dict]) -> Union[None, Response]:
        if not isinstance(payload, list):
            raise ClickhouseError("Payload must be list")

        _tags = [f"database:{database}", f"table:{table}"]
        data = " ".join([json.dumps(row) for row in payload])
        logger.debug(f"Inserting batch: {data}")
        query_result = self.execute(f"INSERT INTO {database}.{table} FORMAT JSONEachRow {data}")
        monitoring.send_gauge_metric("clickhouse_inserted_rows", len(payload), _tags)
        return query_result

    @monitoring.send_time_metric("clickhouse_deduplicate_time_seconds")
    def deduplicate(self, database: str, table: str) -> Union[None, Response]:
        return self.execute(f"OPTIMIZE TABLE {database}.{table} FINAL")



import logging

from typing import List, Tuple
from yandex_tracker_client import TrackerClient
from yandex_tracker_client.collections import Issues, IssueComments

from tracker_exporter.models.issue import TrackerIssue
from tracker_exporter.models.enums import YandexTrackerLanguages
from tracker_exporter.utils.helpers import fix_null_dates
from tracker_exporter.services.monitoring import DogStatsdClient
from tracker_exporter.errors import TrackerError

logger = logging.getLogger(__name__)
monitoring = DogStatsdClient()


class YandexTrackerClient:
    """This class provide simple facade interface for Yandex.Tracker."""

    def __init__(self,
                 org_id: str,
                 token: str,
                 lang: YandexTrackerLanguages = YandexTrackerLanguages.EN):

        self.token = token
        self.org_id = str(org_id)
        self.lang = lang

        if self.lang.lower() not in ("en", "ru"):
            raise TrackerError("Tracker client language must be 'en' or 'ru'")

        self.client = TrackerClient(
            token=self.token,
            org_id=self.org_id,
            headers={"Accept-Language": self.lang}
        )

    def get_issue(self, issue_key: str) -> Issues:
        return self.client.issues[issue_key]

    def get_comments(self, issue_key: str) -> IssueComments:
        return self.client.issues[issue_key].comments.get_all()

    @monitoring.send_time_metric("issues_search_time_seconds")
    def search_issues(self, query: str, limit: int = 100) -> List[Issues]:
        found_issues = self.client.issues.find(query, per_page=limit)
        logger.info(f"Found {len(found_issues)} issues by query '{query}'")
        return found_issues

    @monitoring.send_time_metric("issue_transform_time_seconds")
    def issue_cycle_time(self, issue_key: str) -> Tuple[List[dict]]:
        issue = TrackerIssue(self.get_issue(issue_key))
        metrics = issue.metrics()

        if not metrics:
            return fix_null_dates(issue.to_dict()), None
        return fix_null_dates(issue.to_dict()), [m.to_dict() for m in metrics]



import os
import datetime as dt

# Common settings
LOGLEVEL = os.environ.get("EXPORTER_LOGLEVEL", "info")
UPLOAD_TO_STORAGE = os.environ.get("EXPORTER_ENABLE_UPLOAD", "false").lower() in ("true", "yes")

# Business days settings
WORKDAYS = [0, 1, 2, 3, 4]  # From Monday to Friday
try:
    BUSINESS_HOURS_START = dt.time(int(os.environ.get("EXPORTER_WORKHOURS_START", "9")))
    BUSINESS_HOURS_END = dt.time(int(os.environ.get("EXPORTER_WORKHOURS_END", "22")))
except (ValueError, TypeError):
    BUSINESS_HOURS_START = dt.time(9)
    BUSINESS_HOURS_END = dt.time(22)

# Monitoring settings
MONITORING_ENABLED = os.environ.get("EXPORTER_MONITORING_ENABLED", "false").lower() in ("true", "yes")
MONITORING_HOST = os.environ.get("EXPORTER_MONITORING_HOST", "localhost")
MONITORING_PORT = os.environ.get("EXPORTER_MONITORING_PORT", 8125)
MONITORING_METRIC_BASE_PREFIX = os.environ.get("MONITORING_METRIC_PREFIX", "tracker_exporter")
MONITORING_BASE_LABELS = [
    "project:internal",
]
SENTRY_ENABLED = os.environ.get("EXPORTER_SENTRY_ENABLED", "false").lower() in ("true", "yes")
SENTRY_DSN = os.environ.get("EXPORTER_SENTRY_DSN")

# Tracker settings
TRACKER_DEFAULT_DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
TRACKER_BULK_CYCLE_TIME_ISSUES_LIMIT = 1000
TRACKER_TOKEN = os.environ.get("EXPORTER_TRACKER_TOKEN")
TRACKER_ORG_ID = os.environ.get("EXPORTER_TRACKER_ORG_ID")
TRACKER_ISSUES_UPDATE_INTERVAL = os.environ.get("EXPORTER_TRACKER_ISSUES_FETCH_INTERVAL", 30)  # min
TRACKER_ISSUES_SEARCH_QUERY = os.environ.get("EXPORTER_TRACKER_ISSUES_SEARCH_QUERY")
TRACKER_ISSUES_SEARCH_RANGE = os.environ.get("EXPORTER_TRACKER_ISSUES_SEARCH_RANGE", "2h")

# Clickhouse settings
CLICKHOUSE_HOST = os.environ.get("EXPORTER_CLICKHOUSE_HOST", "localhost")
CLICKHOUSE_PROTO = os.environ.get("EXPORTER_CLICKHOUSE_PROTO", "http")
CLICKHOUSE_HTTP_PORT = os.environ.get("EXPORTER_CLICKHOUSE_HTTP_PORT", 8123)
CLICKHOUSE_CACERT_PATH = os.environ.get("EXPORTER_CLICKHOUSE_CERT", None)
CLICKHOUSE_SERVERLESS_PROXY_ID = os.environ.get("EXPORTER_CLICKHOUSE_SERVERLESS_PROXY_ID", None)
CLICKHOUSE_USER = os.environ.get("EXPORTER_CLICKHOUSE_USER", "default")
CLICKHOUSE_PASSWORD = os.environ.get("EXPORTER_CLICKHOUSE_PASSWORD")
CLICKHOUSE_DATABASE = os.environ.get("EXPORTER_CLICKHOUSE_DATABASE", "agile")
CLICKHOUSE_ISSUES_TABLE = os.environ.get("EXPORTER_CLICKHOUSE_ISSUES_TABLE", "issues")
CLICKHOUSE_ISSUE_METRICS_TABLE = os.environ.get("EXPORTER_CLICKHOUSE_ISSUE_METRICS_TABLE", "issue_metrics")

# Exporter settings
_DEFAULT_CLOSED_ISSUE_STATUSES = "closed,rejected,resolved,cancelled,released"
CLOSED_ISSUE_STATUSES = os.environ.get("EXPORTER_CLOSED_ISSUE_STATUES", _DEFAULT_CLOSED_ISSUE_STATUSES).split(",")
EXCLUDE_QUEUES = (
    "TEST",
)
NOT_NULLABLE_FIELDS = (
    "created_at",
    "resolved_at",
    "closed_at",
    "updated_at",
    "released_at",
    "deadline",
    "start_date",
    "end_date",
    "start_time",
    "end_time",
)



class TrackerChangelogEvents:
    ISSUE_WORKFLOW = "IssueWorkflow"


class TrackerWorkflowTypes:
    TRANSITION = "status"
    RESOLVE_ISSUE = "resolution"


class YandexTrackerLanguages:
    RU = "ru"
    EN = "en"


class TimeDeltaOut:
    SECONDS = "seconds"
    MINUTES = "minutes"

class ClickhouseProto:
    HTTPS = "https"
    HTTP = "http"



# pylint: disable=W0102
import logging

from functools import wraps
from datadog import DogStatsd

from yandex_tracker_client.exceptions import (
    TrackerError,
    TrackerServerError,
    TrackerRequestError,
    TrackerClientError,
)
from tracker_exporter.defaults import (
    MONITORING_METRIC_BASE_PREFIX,
    MONITORING_HOST,
    MONITORING_PORT,
    MONITORING_BASE_LABELS,
    MONITORING_ENABLED
)

logger = logging.getLogger(__name__)


class DogStatsdClient:
    """This class represents interface for DataDog statsd UDP client."""

    def __init__(self,
                 host: str = MONITORING_HOST,
                 port: int = MONITORING_PORT,
                 base_labels: list = MONITORING_BASE_LABELS,  # pylint: disable=W0102
                 metric_name_prefix: str = MONITORING_METRIC_BASE_PREFIX,
                 use_ms: bool = True,
                 enabled: bool = MONITORING_ENABLED):

        self.host = host
        self.port = int(port)
        self.base_labels = base_labels
        self.prefix = metric_name_prefix
        self._enabled = enabled
        self.client = DogStatsd(
            host=self.host,
            port=self.port,
            use_ms=use_ms,
            constant_tags=self.base_labels
        )
        if self._enabled:
            assert self.host is not None
            assert self.port is not None
        logger.info(f"Monitoring send metrics is {'enabled' if self._enabled else 'disabled'}")

    def send_count_metric(self,
                          name: str,
                          value: int,
                          tags: list = []):
        metric = f"{self.prefix}_{name}"
        def metric_wrapper(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if not self._enabled:
                    return func(*args, **kwargs)
                self.client.increment(metric, value, tags=tags)
                logger.debug(f"Success sent count metric: {metric}")
                return func(*args, **kwargs)
            return wrapper
        return metric_wrapper

    def send_gauge_metric(self,
                          name: str,
                          value: int,
                          tags: list = []):
        if not self._enabled:
            return
        metric = f"{self.prefix}_{name}"
        self.client.gauge(metric, value, tags=tags)
        logger.debug(f"Success sent gauge metric: {metric}")

    def send_time_metric(self,
                         name: str,
                         tags: list = []):
        metric = f"{self.prefix}_{name}"
        def metric_wrapper(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if not self._enabled:
                    return func(*args, **kwargs)
                with self.client.timed(metric, tags=tags):
                    return func(*args, **kwargs)
            logger.debug(f"Success start time metric: {metric}")
            return wrapper
        return metric_wrapper


def sentry_events_filter(event, hint):  # pylint: disable=R1710
    #  Drop all events without exception trace
    if "exc_info" not in hint:
        return

    exception = hint["exc_info"][1]
    if isinstance(exception, (TrackerError, TrackerClientError, TrackerRequestError, TrackerServerError)):
        event["fingerprint"] = ["tracker-error"]

    return event



# pylint: disable=C0103,W0622
version = "0.1.18"
url = "https://github.com/akimrx/yandex-tracker-exporter"
download_url = "https://pypi.org/project/tracker-exporter/"
appname = "yandex_tracker_exporter"
description = "Yandex.Tracker issue metrics exporter"
author = "Akim Faskhutdinov"
author_email = "akimstrong@yandex.ru"
license = "MIT"



import json
from abc import ABCMeta


class Base:
    """Base class for objects."""

    __metaclass__ = ABCMeta

    def __str__(self) -> str:
        return str(self.to_dict())

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, item):
        return self.__dict__[item]

    @classmethod
    def de_json(cls, data) -> dict:
        """Deserialize object."""
        if not data:
            return None

        data = data.copy()
        return data

    def to_json(self) -> dict:
        """Serialize object to json."""
        return json.dumps(self.to_dict())

    def to_dict(self) -> dict:
        """Recursive serialize object."""
        null_cleaner = lambda value: "" if value is None else value  # pylint: disable=C3001

        def parse(val):
            if isinstance(val, list):
                return [parse(it) for it in val]
            if isinstance(val, dict):
                return {key: null_cleaner(parse(value)) for key, value in val.items() if not key.startswith("_")}
            return val

        data = self.__dict__.copy()
        return parse(data)



#!/usr/bin/env python3

import os
import sys
import time
import signal
import logging
import warnings
import argparse

from datetime import datetime, timedelta
from dotenv import load_dotenv

import sentry_sdk
from apscheduler.schedulers.background import BackgroundScheduler

parser = argparse.ArgumentParser("tracker-exporter")
parser.add_argument(
    "-e", "--env-file",
    metavar="file",
    dest="env_file",
    type=str,
    required=False,
    help="Path to .env file"
)
args = parser.parse_args()
warnings.filterwarnings("ignore")

if args.env_file:
    load_dotenv(args.env_file)


# pylint: disable=C0413
from .errors import ExportError
from .services.monitoring import DogStatsdClient, sentry_events_filter
from .exporter import Exporter
from .__version__ import appname, version
from .defaults import (
    EXCLUDE_QUEUES,
    LOGLEVEL,
    UPLOAD_TO_STORAGE,
    TRACKER_ISSUES_UPDATE_INTERVAL,
    SENTRY_ENABLED,
    SENTRY_DSN,
)

logging.basicConfig(
    level=LOGLEVEL.upper(),
    datefmt="%Y-%m-%d %H:%M:%S",
    format="%(asctime)s [%(levelname)s] [%(name)s.%(funcName)s] %(message)s"
)
logging.getLogger("yandex_tracker_client").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)
scheduler = BackgroundScheduler()
monitoring = DogStatsdClient()
exporter = Exporter()
logger.debug(f"Environment: {os.environ.items()}")


def signal_handler(sig, frame) -> None:  # pylint: disable=W0613
    if sig == signal.SIGINT:
        logger.warning("Received SIGINT, graceful shutdown...")
        scheduler.shutdown()
        sys.exit(0)


def configure_sentry() -> None:
    if SENTRY_ENABLED:
        assert SENTRY_DSN is not None
        sentry_sdk.init(
            dsn=SENTRY_DSN,
            traces_sample_rate=1.0,
            release=f"{appname}@{version}",
            before_send=sentry_events_filter
        )
    logger.info(f"Sentry send traces is {'enabled' if SENTRY_ENABLED else 'disabled'}")


def export_cycle_time(exclude_queues: str = EXCLUDE_QUEUES,
                      upload: bool = UPLOAD_TO_STORAGE,
                      ignore_exceptions: bool = True) -> None:
    try:
        inserted_rows = exporter.cycle_time(exclude_queues=exclude_queues, upload=upload)
        if inserted_rows > 0:
            monitoring.send_gauge_metric("last_update_timestamp", value=int(time.time()))
            monitoring.send_gauge_metric("upload_status", value=1)
    except Exception as exc:
        monitoring.send_gauge_metric("upload_status", value=2)
        logger.exception(f"Something error occured: {exc}")
        if not ignore_exceptions:
            raise ExportError(exc) from exc


def main() -> None:
    configure_sentry()
    signal.signal(signal.SIGINT, signal_handler)
    scheduler.start()
    scheduler.add_job(
        export_cycle_time,
        trigger="interval",
        name="issues_cycle_time_exporter",
        minutes=int(TRACKER_ISSUES_UPDATE_INTERVAL),
        max_instances=1,
        next_run_time=datetime.now() + timedelta(seconds=5)
    )
    signal.pause()


if __name__ == "__main__":
    main()



import re
import time
import logging

from functools import wraps
from typing import Union, Tuple
from datetime import datetime

import holidays
import pandas as pd
import businesstimedelta

from tracker_exporter.models.enums import TimeDeltaOut
from tracker_exporter.defaults import (
    NOT_NULLABLE_FIELDS,
    WORKDAYS,
    BUSINESS_HOURS_START,
    BUSINESS_HOURS_END,
    TRACKER_DEFAULT_DATETIME_FORMAT
)

logger = logging.getLogger(__name__)


def get_timedelta(end_time: datetime,
                  start_time: datetime,
                  out: TimeDeltaOut = TimeDeltaOut.SECONDS) -> int:
    """Simple timedelta between dates."""
    assert isinstance(start_time, datetime)
    assert isinstance(end_time, datetime)

    delta = int((end_time - start_time).total_seconds())
    if out == TimeDeltaOut.MINUTES:
        return delta // 60
    if out == TimeDeltaOut.SECONDS:
        return delta
    return delta


def calculate_time_spent(start_date: datetime,
                         end_date: datetime,
                         busdays_only: bool = False,
                         workdays: list = WORKDAYS,
                         business_hours: Tuple = (BUSINESS_HOURS_START, BUSINESS_HOURS_END,)) -> int:
    """
    Calculate timedelta between dates with business days support.
    Weekdays: Monday is 0, Sunday is 6, so weekends (5, 6) mean (Sat, Sun).
    Returns: seconds
    """
    if not isinstance(start_date, datetime):
        start_date = pd.to_datetime(start_date)
    if not isinstance(end_date, datetime):
        end_date = pd.to_datetime(end_date)

    holiday_rules = businesstimedelta.HolidayRule(holidays.RU())
    workday_rules = businesstimedelta.WorkDayRule(
        start_time=business_hours[0],
        end_time=business_hours[1],
        working_days=workdays)


    if busdays_only:
        logger.debug(f"Calculating workhours. Business hours: {business_hours}. {start_date}, {end_date}")
        bt = businesstimedelta.Rules([workday_rules, holiday_rules])
        result = bt.difference(start_date, end_date).timedelta.total_seconds()
    else:
        logger.debug(f"Calculating regular hours")
        result = (end_date - start_date).total_seconds()

    return abs(int(result))



def fix_null_dates(data: dict) -> dict:
    to_remove = []

    for key, value in data.items():
        if key in NOT_NULLABLE_FIELDS and (value is None or value == ""):
            to_remove.append(key)

    for key in to_remove:
        del data[key]

    return data


# pylint: disable=R1710
def validate_resource(resource: object,
                      attribute: str,
                      low: bool = True) -> Union[str, list, bool, int, None]:
    """Validate Yandex.Tracker object attribute and return it if exists as string."""
    if hasattr(resource, attribute):
        _attr = getattr(resource, attribute)
        if isinstance(_attr, str):
            if low:
                return _attr.lower()
            return _attr
        return _attr


def to_snake_case(text: str) -> str:
    """Convert any string to `snake_case` format."""
    if text is None or text == "":
        return text

    text = re.sub(r"('|\")", "", text)
    string = re.sub(r"(_|-)+", " ", text).title().replace(" ", "")
    output = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', output).lower()


def to_simple_datetime(dtime: str,
                       source_dt_format: str = TRACKER_DEFAULT_DATETIME_FORMAT,
                       date_only: bool = False,
                       shift: int = 3) -> str:
    """Return (Unicode) date format `YYYY-MM-DD HH:mm:ss` or `YYYY-MM-DD` if `date_only`."""
    if dtime is None:
        logger.debug("dtime is empty, can't transform date to simple string.")
        return

    if date_only:
        fmt = "%Y-%m-%d"
    else:
        fmt = "%Y-%m-%d %H:%M:%S"

    timestamp = time.mktime(datetime.strptime(dtime.split(".")[0], source_dt_format).timetuple())
    if shift > 0:
        timestamp += 60 * 60 * shift
    elif shift < 0:
        timestamp -= 60 * 60 * shift
    return datetime.fromtimestamp(int(timestamp)).strftime(fmt)


def retry(exceptions: tuple, tries: int = 3, delay: Union[float, int] = 1, backoff: int = 3):
    """Decorator to retry the execution of the func, if you have received the errors listed."""
    def retry_decorator(func):
        @wraps(func)
        def func_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            counter = 0
            while mtries > 0:
                try:
                    counter += 1
                    return func(*args, **kwargs)
                except exceptions as err:
                    mtries -= 1
                    if mtries == 0:
                        logger.warning(f"{func.__name__} has failed {counter} times")
                        raise err
                    logger.warning(
                        f"Error in func {func.__name__}, cause: {err}. "
                        f"Retrying ({counter}) in {mdelay} seconds..."
                    )
                    time.sleep(mdelay)
                    mdelay *= backoff
        return func_retry
    return retry_decorator


def to_human_time(seconds: Union[int, float], verbosity: int = 2) -> str:
    """Convert seconds to human readable timedelta
    like a `2w 3d 1h 20m`
    """
    seconds = int(seconds)
    if seconds == 0:
        return "0s"

    negative = False
    if seconds < 0:
        negative = True
        seconds = abs(seconds)

    result = []
    intervals = (
        ("y", 31104000),
        ("mo", 2592000),
        ("w", 604800),
        ("d", 86400),
        ("h", 3600),
        ("m", 60),
        ("s", 1),
    )
    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            result.append(f"{value}{name}")
    delta = " ".join(result[:verbosity])
    return f"-{delta}" if negative else delta


def string_normalize(text: str) -> str:
    """Remove all incompatible symbols."""
    emoji_pattern = re.compile("["
            u"\U0001F600-\U0001F64F"  # emoticons
            u"\U0001F300-\U0001F5FF"  # symbols & pictographs
            u"\U0001F680-\U0001F6FF"  # transport & map symbols
            u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                            "]+", flags=re.UNICODE)
    return emoji_pattern.sub(r"", text)



#!/usr/bin/env python3

from os import path
from setuptools import find_packages, setup


def readme():
    with open("README.md", "r") as fh:
        long_description = fh.read()

    return long_description

cwd = path.abspath(path.dirname(__file__))


def metadata():
    meta = {}
    with open(path.join(cwd, "tracker_exporter", "__version__.py"), "r") as fh:
        exec(fh.read(), meta)
    return meta


def requirements():
    requirements_list = []

    with open("requirements.txt") as requirements:
        for install in requirements:
            requirements_list.append(install.strip())

    return requirements_list

metadata = metadata()
readme = readme()
packages = find_packages()
requirements = requirements()


def main():
    setup(
        name="tracker-exporter",
        version=metadata.get("version"),
        author=metadata.get("author"),
        author_email=metadata.get("author_email"),
        license=metadata.get("license"),
        description=metadata.get("description"),
        long_description=readme,
        long_description_content_type="text/markdown",
        url=metadata.get("url"),
        download_url=metadata.get("download_url"),
        keywords=["yandex tracker exporter", "agile", "cycle time"],
        platforms=["osx", "linux"],
        packages=packages,
        classifiers = [
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
        ],
        install_requires=requirements,
        include_package_data=True,
        python_requires=">=3.7",
        entry_points={
            "console_scripts": [
                "tracker-exporter=tracker_exporter.main:main"
            ]
        },
        zip_safe=False
    )


if __name__ == "__main__":
    main()